@@ -11,9 +11,7 @@
  * was very easy to do with just the text editor.
  */
 
-#include "charmap.h"
-#include "utf8.h"
-#include "i18n.h"
+#include "types.h"
 
 /*
  * This is an implementation of wcwidth() and wcswidth() (defined in
@@ -149,8 +147,6 @@ static int bisearch(int ucs, const struct interval *table, int max)
 
 int joe_wcwidth(int wide,int ucs)
 {
-	int x;
-
 	/* sorted list of non-overlapping intervals of non-spacing characters */
 	/* generated by "uniset +cat=Me +cat=Mn +cat=Cf -00AD +1160-11FF +200B c" */
 	static const struct interval combining[] = {
@@ -828,7 +824,7 @@ int joe_iswalpha_(struct charmap *foo,int c)
 
 int joe_iswalnum_(struct charmap *foo,int c)
 {
-	if (c>=0x30 && c<=0x39 || c==0x5F)
+	if ((c>=0x30 && c<=0x39) || c==0x5F)
 		return 1;
 	else
 		return joe_iswalpha(foo,c);
