@@ -5,78 +5,130 @@
  *
  *	This file is part of JOE (Joe's Own Editor)
  */
-#include "config.h"
 #include "types.h"
 
-#include "scrn.h"
-#include "utils.h"
-#include "vs.h"
-#include "charmap.h"
-#include "utf8.h"
-#include "w.h"
+/* Return width of a string */
 
-extern int bg_prompt;
+int joe_wcswidth(struct charmap *map,unsigned char *s, int len)
+{
+	if (!map->type) {
+		return len;
+	} else {
+		int width = 0;
+		while (len) {
+			int c = utf8_decode_fwrd(&s, &len);
+			if (c >= 0) {
+				width += joe_wcwidth(1, c);
+			} else
+				++width;
+		}
+		return width;
+	}
+}
+
+/* Calculate number of lines needed for a given prompt string and a given window width.
+   Also this finds the nth line and returns the position of the substring which is
+   that line. Set n to -1 if you just want the height. */
+
+int break_height(struct charmap *map,unsigned char **src,int *src_len,int wid,int n)
+{
+	unsigned char *s = *src;
+	int len = *src_len;
+	int h = 1; /* Number of lines */
+	int col = 0; /* Current column */
+	int x = 0; /* Offset into string */
+	int start_of_line = 0; /* Start of most recent line */
+	while (x != len) {
+		int space = 0;
+		int word = 0;
+		int start = x;
+		int start_word;
+		while (x != len && s[x] == ' ') {
+			++space;
+			++x;
+		}
+		start_word = x;
+		while (x != len && s[x] != ' ') {
+			++x;
+		}
+		word = joe_wcswidth(map, s + start_word, x - start_word);
+		if (col + space + word < wid || !col) {
+			/* Leading space and word fit on current line */
+			col += space + word;
+		} else {
+			/* They don't fit, start a new line */
+			if (!n--) {
+				x = start;
+				break;
+			}
+			++h;
+			col = word;
+			start_of_line = start_word;
+		}
+	}
+	*src = s + start_of_line;
+	*src_len = x - start_of_line;
+	return h;
+}
 
 static void dispqw(QW *qw)
 {
+	int y;
 	W *w = qw->parent;
 
-	/* Scroll buffer and position prompt */
-	if (qw->promptlen > w->w / 2 + w->w / 4)
-		qw->promptofst = qw->promptlen - w->w / 2;
-	else
-		qw->promptofst = 0;
-
-	/* Set cursor position */
-	w->curx = qw->promptlen - qw->promptofst;
-	w->cury = 0;
-
 	/* Generate prompt */
-	w->t->t->updtab[w->y] = 1;
-	genfield(w->t->t,
-	         w->t->t->scrn + w->y * w->t->t->co + w->x,
-	         w->t->t->attr + w->y * w->t->t->co + w->x,
-	         w->x,
-	         w->y,
-	         qw->promptofst,
-	         qw->prompt,
-	         qw->promptlen,
-	         BG_COLOR(bg_prompt),
-	         w->w - w->x,
-	         1,NULL);
+	for (y = 0; y != w->h; ++y) {
+		unsigned char *s = qw->prompt;
+		int l = qw->promptlen;
+		break_height(locale_map, &s, &l, qw->org_w, y);
+		w->t->t->updtab[w->y + y] = 1;
+		genfield(w->t->t,
+		         w->t->t->scrn + (w->y + y) * w->t->t->co + w->x,
+		         w->t->t->attr + (w->y + y) * w->t->t->co + w->x,
+		         w->x,
+		         w->y + y,
+		         0,
+		         s,
+		         l,
+		         BG_COLOR(bg_prompt),
+		         w->w - w->x,
+		         1,NULL);
+		w->cury = y;
+		w->curx = w->x + joe_wcswidth(locale_map, s, l);
+	}
 }
 
 static void dispqwn(QW *qw)
 {
+	int y;
 	W *w = qw->parent;
 
-	/* Scroll buffer and position prompt */
-	if (qw->promptlen > w->w / 2 + w->w / 4)
-		qw->promptofst = qw->promptlen - w->w / 2;
-	else
-		qw->promptofst = 0;
-
 	/* Set cursor position */
 	if (w->win->watom->follow && w->win->object)
 		w->win->watom->follow(w->win->object);
 	if (w->win->watom->disp && w->win->object)
-		w->win->watom->disp(w->win->object);
+		w->win->watom->disp(w->win->object, 1);
 	w->curx = w->win->curx;
 	w->cury = w->win->cury + w->win->y - w->y;
 
 	/* Generate prompt */
-	w->t->t->updtab[w->y] = 1;
-	genfield(w->t->t,
-	         w->t->t->scrn + w->y * w->t->t->co + w->x,
-	         w->t->t->attr + w->y * w->t->t->co + w->x,
-	         w->x,
-	         w->y,
-	         qw->promptofst,
-	         qw->prompt,
-	         qw->promptlen,
-	         BG_COLOR(bg_prompt),
-	         w->w - w->x,
-	         1,NULL);
+	for (y = 0; y != w->h; ++y) {
+		unsigned char *s = qw->prompt;
+		int l = qw->promptlen;
+		break_height(locale_map, &s, &l, qw->org_w, y);
+		w->t->t->updtab[w->y + y] = 1;
+		genfield(w->t->t,
+		         w->t->t->scrn + (w->y + y) * w->t->t->co + w->x,
+		         w->t->t->attr + (w->y + y) * w->t->t->co + w->x,
+		         w->x,
+		         w->y + y,
+		         0,
+		         s,
+		         l,
+		         BG_COLOR(bg_prompt),
+		         w->w - w->x,
+		         1,NULL);
+	}
 }
 
 /* When user hits a key in a query window */
@@ -124,7 +176,7 @@ static int abortqw(QW *qw)
 }
 
 static WATOM watomqw = {
-	US "query",
+	USTR "query",
 	dispqw,
 	NULL,
 	abortqw,
@@ -138,7 +190,7 @@ static WATOM watomqw = {
 };
 
 static WATOM watqwn = {
-	US "querya",
+	USTR "querya",
 	dispqwn,
 	NULL,
 	abortqw,
@@ -152,7 +204,7 @@ static WATOM watqwn = {
 };
 
 static WATOM watqwsr = {
-	US "querysr",
+	USTR "querysr",
 	dispqwn,
 	NULL,
 	abortqw,
@@ -171,8 +223,11 @@ QW *mkqw(W *w, unsigned char *prompt, int len, int (*func) (/* ??? */), int (*ab
 {
 	W *new;
 	QW *qw;
+	unsigned char *s = prompt;
+	int l = len;
+	int h = break_height(locale_map, &s, &l, w->w, -1);
 
-	new = wcreate(w->t, &watomqw, w, w, w->main, 1, NULL, notify);
+	new = wcreate(w->t, &watomqw, w, w, w->main, h, NULL, notify);
 	if (!new) {
 		if (notify)
 			*notify = 1;
@@ -183,7 +238,8 @@ QW *mkqw(W *w, unsigned char *prompt, int len, int (*func) (/* ??? */), int (*ab
 	qw->parent = new;
 	qw->prompt = vsncpy(NULL, 0, prompt, len);
 	qw->promptlen = len;
-	qw->promptofst = 0;
+	qw->org_w = w->w;
+	qw->org_h = h;
 	qw->func = func;
 	qw->abrt = abrt;
 	qw->object = object;
@@ -198,8 +254,11 @@ QW *mkqwna(W *w, unsigned char *prompt, int len, int (*func) (/* ??? */), int (*
 {
 	W *new;
 	QW *qw;
+	unsigned char *s = prompt;
+	int l = len;
+	int h = break_height(locale_map, &s, &l, w->w, -1);
 
-	new = wcreate(w->t, &watqwn, w, w, w->main, 1, NULL, notify);
+	new = wcreate(w->t, &watqwn, w, w, w->main, h, NULL, notify);
 	if (!new) {
 		if (notify)
 			*notify = 1;
@@ -210,7 +269,8 @@ QW *mkqwna(W *w, unsigned char *prompt, int len, int (*func) (/* ??? */), int (*
 	qw->parent = new;
 	qw->prompt = vsncpy(NULL, 0, prompt, len);
 	qw->promptlen = len;
-	qw->promptofst = 0;
+	qw->org_w = w->w;
+	qw->org_h = h;
 	qw->func = func;
 	qw->abrt = abrt;
 	qw->object = object;
@@ -225,8 +285,11 @@ QW *mkqwnsr(W *w, unsigned char *prompt, int len, int (*func) (/* ??? */), int (
 {
 	W *new;
 	QW *qw;
+	unsigned char *s = prompt;
+	int l = len;
+	int h = break_height(locale_map, &s, &l, w->w, -1);
 
-	new = wcreate(w->t, &watqwsr, w, w, w->main, 1, NULL, notify);
+	new = wcreate(w->t, &watqwsr, w, w, w->main, h, NULL, notify);
 	if (!new) {
 		if (notify)
 			*notify = 1;
@@ -237,7 +300,8 @@ QW *mkqwnsr(W *w, unsigned char *prompt, int len, int (*func) (/* ??? */), int (
 	qw->parent = new;
 	qw->prompt = vsncpy(NULL, 0, prompt, len);
 	qw->promptlen = len;
-	qw->promptofst = 0;
+	qw->org_w = w->w;
+	qw->org_h = h;
 	qw->func = func;
 	qw->abrt = abrt;
 	qw->object = object;
