@@ -8,9 +8,6 @@
 #ifndef _JOE_RC_H
 #define _JOE_RC_H 1
 
-#include "config.h"
-#include "types.h"
-
 extern OPTIONS pdefault;
 extern OPTIONS fdefault;
 void setopt PARAMS((B *b, unsigned char *name));
@@ -39,6 +36,7 @@ int procrc PARAMS((CAP *cap, unsigned char *name));
 int glopt PARAMS((unsigned char *s, unsigned char *arg, OPTIONS *options, int set));
 
 int umode PARAMS((BW *bw));
+int umenu PARAMS((BW *bw));
 
 /* Save state */
 void save_state();
@@ -46,4 +44,15 @@ void save_state();
 /* Load state */
 void load_state();
 
+/* Validate rc file: return -1 if it's bad (call this after rc file has been loaded) */
+int validate_rc();
+
+/* Update options */
+void lazy_opts PARAMS((B *b, OPTIONS *o));
+
+int ucharset PARAMS((BW *bw));
+int ulanguage PARAMS((BW *bw));
+
+unsigned char *get_status(BW *bw, unsigned char *s);
+
 #endif
