@@ -8,9 +8,8 @@
 #ifndef _JOE_TAB_H
 #define _JOE_TAB_H 1
 
-#include "config.h"
-#include "types.h"
-
 int cmplt PARAMS((BW *bw));
+extern int menu_jump;
+extern int menu_explorer;
 
 #endif
