@@ -5,29 +5,20 @@
  *
  *	This file is part of JOE (Joe's Own Editor)
  */
-#include "config.h"
 #include "types.h"
 
-#include <stdio.h>
-#ifdef HAVE_SYS_TYPES_H
-#include <sys/types.h>
-#endif
-#ifdef HAVE_SYS_STAT_H
-#include <sys/stat.h>
-#endif
-#ifdef HAVE_STDLIB_H
-#include <stdlib.h>
-#endif
 #ifdef TERMINFO
-#ifdef HAVE_TERM_H
+
+#ifdef HAVE_CURSES_H
+#include <curses.h>
 #endif
+/* curses has to come before term.h on SGI */
+#ifdef HAVE_TERM_H
+/* term.h is a disaster: it #defines 'tab' */
+#include <term.h>
 #endif
 
-#include "blocks.h"
-#include "termcap.h"
-#include "utils.h"
-#include "va.h"
-#include "vs.h"
+#endif
 
 int dopadding = 0;
 unsigned char *joeterm = NULL;
@@ -134,7 +125,7 @@ static long findidx(FILE *file, unsigned char *name)
 			buf[y] = 0;
 			if (c == '\n' || !c) {
 				z = 0;
-				sscanf((char *)(buf + x), "%x", &z);
+				sscanf((char *)(buf + x), "%x", (unsigned *)&z);
 				addr += z;
 			} else if (!zcmp(buf + x, name))
 				flg = 1;
@@ -148,7 +139,7 @@ static long findidx(FILE *file, unsigned char *name)
 
 /* Load termcap entry */
 
-CAP *getcap(unsigned char *name, unsigned int baud, void (*out) (unsigned char *, unsigned char), void *outptr)
+CAP *my_getcap(unsigned char *name, unsigned int baud, void (*out) (unsigned char *, unsigned char), void *outptr)
 {
 	CAP *cap;
 	FILE *f, *f1;
@@ -222,7 +213,7 @@ CAP *getcap(unsigned char *name, unsigned int baud, void (*out) (unsigned char *
  joe_free(cap);
  return 0;
 */
-		fprintf(stderr, "Couldn't load termcap entry.  Using ansi default\n");
+		fprintf(stderr, (char *)joe_gettext(_("Couldn't load termcap entry.  Using ansi default\n")));
 		ti = 0;
 		cap->tbuf = vsncpy(cap->tbuf, 0, sc(defentry));
 		goto checktc;
@@ -243,7 +234,7 @@ CAP *getcap(unsigned char *name, unsigned int baud, void (*out) (unsigned char *
 		if (buf.st_mtime > buf1.st_mtime)
 			idx = findidx(f, name);
 		else
-			fprintf(stderr, "%s is out of date\n", idxname);
+			fprintf(stderr, (char *)joe_gettext(_("%s is out of date\n")), idxname);
 		fclose(f);
 	}
 	vsrm(idxname);
@@ -352,7 +343,7 @@ CAP *getcap(unsigned char *name, unsigned int baud, void (*out) (unsigned char *
 	varm(npbuf);
 	vsrm(name);
 
-	cap->pad = jgetstr(cap, US "pc");
+	cap->pad = jgetstr(cap, USTR "pc");
 	if (dopadding)
 		cap->dopadding = 1;
 	else
@@ -409,8 +400,13 @@ unsigned char *jgetstr(CAP *cap, unsigned char *name)
 	struct sortentry *s;
 
 #ifdef TERMINFO
-	if (cap->abuf)
-		return (unsigned char *)tgetstr((char *)name, (char **)&cap->abufp);
+	if (cap->abuf) {
+		char *new_ptr = (char *)cap->abufp;
+		char *rtn;
+		rtn = tgetstr((char *)name, &new_ptr);
+		cap->abufp = (unsigned char *)new_ptr;
+		return (unsigned char *)rtn;
+	}
 #endif
 	s = findcap(cap, name);
 	if (s)
@@ -443,7 +439,7 @@ void rmcap(CAP *cap)
 	joe_free(cap);
 }
 
-static unsigned char escape(unsigned char **s)
+static unsigned char escape1(unsigned char **s)
 {
 	unsigned char c = *(*s)++;
 
@@ -545,7 +541,7 @@ void texec(CAP *cap, unsigned char *s, int l, int a0, int a1, int a2, int a3)
 /* Output string */
 	while ((c = *s++) != '\0')
 		if (c == '%' && *s) {
-			switch (x = a[0], c = escape(&s)) {
+			switch (x = a[0], c = escape1(&s)) {
 			case 'C':
 				if (x >= 96) {
 					cap->out(cap->outptr, x / 96);
@@ -553,7 +549,7 @@ void texec(CAP *cap, unsigned char *s, int l, int a0, int a1, int a2, int a3)
 				}
 			case '+':
 				if (*s)
-					x += escape(&s);
+					x += escape1(&s);
 			case '.':
 				cap->out(cap->outptr, x);
 				++a;
@@ -640,17 +636,17 @@ void texec(CAP *cap, unsigned char *s, int l, int a0, int a1, int a2, int a3)
 				a[0] = 16 * (a[0] / 10) + a[0] % 10;
 				break;
 			case '>':
-				if (a[0] > escape(&s))
-					a[0] += escape(&s);
+				if (a[0] > escape1(&s))
+					a[0] += escape1(&s);
 				else
-					escape(&s);
+					escape1(&s);
 			default:
 				cap->out(cap->outptr, '%');
 				cap->out(cap->outptr, c);
 			}
 		} else {
 			--s;
-			cap->out(cap->outptr, escape(&s));
+			cap->out(cap->outptr, escape1(&s));
 		}
 
 /* Output padding characters */
@@ -727,7 +723,7 @@ unsigned char c;
 int tgetent(buf, name)
 unsigned char *buf, *name;
 {
-	latest = getcap(name, 9600, stupid, NULL);
+	latest = my_getcap(name, 9600, stupid, NULL);
 	if (latest)
 		return 1;
 	else
