@@ -5,22 +5,16 @@
  *
  *	This file is part of JOE (Joe's Own Editor)
  */
-#include "config.h"
 #include "types.h"
 
-#include <sys/types.h>
-#ifdef HAVE_SYS_STAT_H
-#include <sys/stat.h>
-#endif
+/* not needed: ioctl() moved to utils.c
 #ifdef GWINSZ_IN_SYS_IOCTL
 #ifdef HAVE_SYS_IOCTL_H
 #include <sys/ioctl.h>
 #endif
 #endif
-#include <stdio.h>
-#ifdef HAVE_FCNTL_H
-#include <fcntl.h>
-#endif
+*/
+
 #ifdef HAVE_SYS_WAIT_H
 #include <sys/wait.h>
 #endif
@@ -29,19 +23,25 @@
 #endif
 
 #ifdef HAVE_OPENPTY
+
 #ifdef HAVE_PTY_H
 #include <pty.h>
 #endif
+
+#ifdef HAVE_LIBUTIL_H
+#include <libutil.h>
 #endif
 
-#include <errno.h>
-#include <stdlib.h>
-#include <unistd.h>
+#endif
+
+#ifdef HAVE_LOGIN_TTY
+#ifdef HAVE_UTMP_H
+#include <utmp.h>
+#endif
+#endif
 
 int idleout = 1;
 
-#include "config.h"
-
 #ifdef __amigaos
 #undef SIGTSTP
 #endif
@@ -52,7 +52,6 @@ int idleout = 1;
  */
 #ifdef HAVE_POSIX_TERMIOS
 #  include <termios.h>
-#  include <sys/termios.h>
 #else
 #  ifdef HAVE_SYSV_TERMIO
 #    include <termio.h>
@@ -90,15 +89,6 @@ int idleout = 1;
 #include <stropts.h>
 #endif
 
-/* JOE include files */
-
-#include "main.h"
-#include "path.h"
-#include "tty.h"
-#include "utils.h"
-#include "mouse.h"
-#include "cmd.h"
-
 /** Aliased defines **/
 
 /* O_NDELAY, O_NONBLOCK, and FNDELAY are all synonyms for placing a descriptor
@@ -201,7 +191,7 @@ static int mpxfd;		/* Editor reads packets from this fd */
 static int mpxsfd;		/* Clients send packets to this fd */
 
 static int nmpx = 0;
-static int accept = NO_MORE_DATA;	/* =-1 if we have last packet */
+static int acceptch = NO_MORE_DATA;	/* =-1 if we have last packet */
 
 struct packet {
 	MPX *who;
@@ -220,6 +210,7 @@ void sigjoe(void)
 	ttysig = 1;
 	joe_set_signal(SIGHUP, ttsig);
 	joe_set_signal(SIGTERM, ttsig);
+	joe_set_signal(SIGABRT, ttsig);
 	joe_set_signal(SIGINT, SIG_IGN);
 	joe_set_signal(SIGPIPE, SIG_IGN);
 }
@@ -230,6 +221,7 @@ void signrm(void)
 	if (!ttysig)
 		return;
 	ttysig = 0;
+	joe_set_signal(SIGABRT, SIG_DFL);
 	joe_set_signal(SIGHUP, SIG_DFL);
 	joe_set_signal(SIGTERM, SIG_DFL);
 	joe_set_signal(SIGINT, SIG_DFL);
@@ -265,14 +257,11 @@ static RETSIGTYPE winchd(int unused)
 /* Second ticker */
 
 int ticked = 0;
-extern int dostaupd;
 static RETSIGTYPE dotick(int unused)
 {
 	ticked = 1;
 }
 
-extern int auto_scroll;
-
 void tickoff(void)
 {
 #ifdef HAVE_SETITIMER
@@ -287,9 +276,6 @@ void tickoff(void)
 #endif
 }
 
-extern int auto_scroll;
-extern int auto_trig_time;
-
 void tickon(void)
 {
 #ifdef HAVE_SETITIMER
@@ -337,7 +323,7 @@ void ttopnn(void)
 
 	if (!termin) {
 		if (idleout ? (!(termin = stdin) || !(termout = stdout)) : (!(termin = fopen("/dev/tty", "r")) || !(termout = fopen("/dev/tty", "w")))) {
-			fprintf(stderr, "Couldn\'t open /dev/tty\n");
+			fprintf(stderr, (char *)joe_gettext(_("Couldn\'t open /dev/tty\n")));
 			exit(1);
 		} else {
 #ifdef SIGWINCH
@@ -366,7 +352,7 @@ void ttopnn(void)
 	bbaud = cfgetospeed(&newterm);
 #else
 #ifdef HAVE_SYSV_TERMIO
-	ioctl(fileno(termin), TCGETA, &oldterm);
+	joe_ioctl(fileno(termin), TCGETA, &oldterm);
 	newterm = oldterm;
 	newterm.c_lflag = 0;
 	if (noxon)
@@ -376,12 +362,12 @@ void ttopnn(void)
 	newterm.c_oflag = 0;
 	newterm.c_cc[VMIN] = 1;
 	newterm.c_cc[VTIME] = 0;
-	ioctl(fileno(termin), TCSETAW, &newterm);
+	joe_ioctl(fileno(termin), TCSETAW, &newterm);
 	bbaud = (newterm.c_cflag & CBAUD);
 #else
-	ioctl(fileno(termin), TIOCGETP, &arg);
-	ioctl(fileno(termin), TIOCGETC, &targ);
-	ioctl(fileno(termin), TIOCGLTC, &ltarg);
+	joe_ioctl(fileno(termin), TIOCGETP, &arg);
+	joe_ioctl(fileno(termin), TIOCGETC, &targ);
+	joe_ioctl(fileno(termin), TIOCGLTC, &ltarg);
 	oarg = arg;
 	otarg = targ;
 	oltarg = ltarg;
@@ -400,9 +386,9 @@ void ttopnn(void)
 	ltarg.t_flushc = -1;
 	ltarg.t_werasc = -1;
 	ltarg.t_lnextc = -1;
-	ioctl(fileno(termin), TIOCSETN, &arg);
-	ioctl(fileno(termin), TIOCSETC, &targ);
-	ioctl(fileno(termin), TIOCSLTC, &ltarg);
+	joe_ioctl(fileno(termin), TIOCSETN, &arg);
+	joe_ioctl(fileno(termin), TIOCSETC, &targ);
+	joe_ioctl(fileno(termin), TIOCSLTC, &ltarg);
 	bbaud = arg.sg_ospeed;
 #endif
 #endif
@@ -451,11 +437,11 @@ void ttclsn(void)
 	tcsetattr(fileno(termin), TCSADRAIN, &oldterm);
 #else
 #ifdef HAVE_SYSV_TERMIO
-	ioctl(fileno(termin), TCSETAW, &oldterm);
+	joe_ioctl(fileno(termin), TCSETAW, &oldterm);
 #else
-	ioctl(fileno(termin), TIOCSETN, &oarg);
-	ioctl(fileno(termin), TIOCSETC, &otarg);
-	ioctl(fileno(termin), TIOCSLTC, &oltarg);
+	joe_ioctl(fileno(termin), TIOCSETN, &oarg);
+	joe_ioctl(fileno(termin), TIOCSETC, &otarg);
+	joe_ioctl(fileno(termin), TIOCSLTC, &oltarg);
 #endif
 #endif
 
@@ -559,14 +545,14 @@ int ttflsh(void)
 	}
 
 	/* Ack previous packet */
-	if (ackkbd != -1 && accept != NO_MORE_DATA && !have) {
+	if (ackkbd != -1 && acceptch != NO_MORE_DATA && !have) {
 		unsigned char c = 0;
 
 		if (pack.who && pack.who->func)
 			joe_write(pack.who->ackfd, &c, 1);
 		else
 			joe_write(ackkbd, &c, 1);
-		accept = NO_MORE_DATA;
+		acceptch = NO_MORE_DATA;
 	}
 
 	/* Check for typeahead or next packet */
@@ -578,7 +564,7 @@ int ttflsh(void)
 				fcntl(mpxfd, F_SETFL, 0);
 				joe_read(mpxfd, pack.data, pack.size);
 				have = 1;
-				accept = pack.ch;
+				acceptch = pack.ch;
 			} else
 				fcntl(mpxfd, F_SETFL, 0);
 		} else {
@@ -649,11 +635,11 @@ int ttgetc(void)
 				else
 					ttsig(0);
 			}
-			accept = pack.ch;
+			acceptch = pack.ch;
 		}
 		have = 0;
 		if (pack.who) {	/* Got bknd input */
-			if (accept != NO_MORE_DATA) {
+			if (acceptch != NO_MORE_DATA) {
 				if (pack.who->func) {
 					pack.who->func(pack.who->object, pack.data, pack.size);
 					edupd(1);
@@ -662,9 +648,9 @@ int ttgetc(void)
 				mpxdied(pack.who);
 			goto loop;
 		} else {
-			if (accept != NO_MORE_DATA) {
+			if (acceptch != NO_MORE_DATA) {
 				tickoff();
-				return accept;
+				return acceptch;
 			}
 			else {
 				tickoff();
@@ -714,13 +700,13 @@ void ttgtsz(int *x, int *y)
 	*y = 0;
 
 #ifdef TIOCGSIZE
-	if (ioctl(fileno(termout), TIOCGSIZE, &getit) != -1) {
+	if (joe_ioctl(fileno(termout), TIOCGSIZE, &getit) != -1) {
 		*x = getit.ts_cols;
 		*y = getit.ts_lines;
 	}
 #else
 #ifdef TIOCGWINSZ
-	if (ioctl(fileno(termout), TIOCGWINSZ, &getit) != -1) {
+	if (joe_ioctl(fileno(termout), TIOCGWINSZ, &getit) != -1) {
 		*x = getit.ws_col;
 		*y = getit.ws_row;
 	}
@@ -735,7 +721,7 @@ int ttshell(unsigned char *cmd)
 	unsigned char *s = (unsigned char *)getenv("SHELL");
 
 	if (!s) {
-		s = US "/bin/sh";
+		s = USTR "/bin/sh";
 		/* return; */
 	}
 	ttclsn();
@@ -754,7 +740,7 @@ int ttshell(unsigned char *cmd)
 		if (cmd)
 			execl((char *)s, (char *)s, "-c", cmd, NULL);
 		else {
-			fprintf(stderr, "You are at the command shell.  Type 'exit' to return\n");
+			fprintf(stderr, (char *)joe_gettext(_("You are at the command shell.  Type 'exit' to return\n")));
 			execl((char *)s, (char *)s, NULL);
 		}
 		_exit(0);
@@ -768,7 +754,7 @@ static void mpxresume(void)
 {
 	int fds[2];
 	pipe(fds);
-	accept = NO_MORE_DATA;
+	acceptch = NO_MORE_DATA;
 	have = 0;
 	if (!(kbdpid = fork())) {
 		close(fds[1]);
@@ -818,7 +804,7 @@ void ttsusp(void)
 	omode = ttymode;
 	mpxsusp();
 	ttclsn();
-	fprintf(stderr, "You have suspended the program.  Type 'fg' to return\n");
+	fprintf(stderr, (char *)joe_gettext(_("You have suspended the program.  Type 'fg' to return\n")));
 	kill(0, SIGTSTP);
 #ifdef junk
 	/* Hmmm... this should not have been necessary */
@@ -894,8 +880,6 @@ extern char *ptsname();
 static unsigned char *getpty(int *ptyfd)
 {
 	int fdm;
-	unsigned char *name;
-
 	*ptyfd = fdm = open("/dev/ptmx", O_RDWR);
 	grantpt(fdm);
 	unlockpt(fdm);
@@ -909,7 +893,6 @@ static unsigned char *getpty(int *ptyfd)
 
 static unsigned char *getpty(int *ptyfd)
 {
-	int fdm;
 	static unsigned char name[32];
 	int ttyfd;
 
@@ -942,13 +925,13 @@ static unsigned char *getpty(int *ptyfd)
 	static unsigned char ttyname[32];
 
 	if (!ptys) {
-		ttydir = US "/dev/pty/";
-		ptydir = US "/dev/ptym/";	/* HPUX systems */
-		if (chpwd(ptydir) || !(ptys = rexpnd(US "pty*")))
+		ttydir = USTR "/dev/pty/";
+		ptydir = USTR "/dev/ptym/";	/* HPUX systems */
+		if (chpwd(ptydir) || !(ptys = rexpnd(USTR "pty*")))
 			if (!ptys) {
-				ttydir = ptydir = US "/dev/";	/* Everyone else */
+				ttydir = ptydir = USTR "/dev/";	/* Everyone else */
 				if (!chpwd(ptydir))
-					ptys = rexpnd(US "pty*");
+					ptys = rexpnd(USTR "pty*");
 			}
 	}
 	chpwd(orgpwd);
@@ -1003,8 +986,6 @@ static RETSIGTYPE death(int unused)
 
 /* Build a new environment, but replace one variable */
 
-extern unsigned char **mainenv;
-
 static unsigned char **newenv(unsigned char **old, unsigned char *s)
 {
 	unsigned char **new;
@@ -1031,14 +1012,16 @@ static unsigned char **newenv(unsigned char **old, unsigned char *s)
 
 /* Create a shell process */
 
-MPX *mpxmk(int *ptyfd, unsigned char *cmd, unsigned char **args, void (*func) (/* ??? */), void *object, void (*die) (/* ??? */), void *dieobj)
+/* If out_only is set, leave program's stdin attached to JOE's stdin */
+
+MPX *mpxmk(int *ptyfd, unsigned char *cmd, unsigned char **args, void (*func) (/* ??? */), void *object, void (*die) (/* ??? */), void *dieobj, int out_only)
 {
 	unsigned char buf[80];
 	int fds[2];
 	int comm[2];
 	pid_t pid;
 	int x;
-	MPX *m;
+	MPX *m = 0;
 	unsigned char *name;
 
 	/* Get pty/tty pair */
@@ -1081,6 +1064,7 @@ MPX *mpxmk(int *ptyfd, unsigned char *cmd, unsigned char **args, void (*func) (/
 	/* PID number pipe */
 	pipe(comm);
 
+
 	/* Create processes... */
 	if (!(m->kpid = fork())) {
 		/* This process copies data from shell to joe */
@@ -1112,7 +1096,7 @@ MPX *mpxmk(int *ptyfd, unsigned char *cmd, unsigned char **args, void (*func) (/
 
 #ifdef TIOCNOTTY
 			x = open("/dev/tty", O_RDWR);
-			ioctl(x, TIOCNOTTY, 0);
+			joe_ioctl(x, TIOCNOTTY, 0);
 #endif
 
 			setsid();	/* I think you do setprgp(0,0) on systems with no setsid() */
@@ -1123,53 +1107,73 @@ MPX *mpxmk(int *ptyfd, unsigned char *cmd, unsigned char **args, void (*func) (/
 #endif
 
 #endif
+
 			/* Close all fds */
-			for (x = 0; x != 32; ++x)
+			for (x = (out_only ? 1 : 0); x != 32; ++x)
 				close(x);	/* Yes, this is quite a kludge... all in the
 						   name of portability */
 
+
 			/* Open the TTY */
 			if ((x = open((char *)name, O_RDWR)) != -1) {	/* Standard input */
-				unsigned char **env = newenv(mainenv, US "TERM=");
+				unsigned char **env = newenv(mainenv, USTR "TERM=");
 
+
+				if (!out_only) {
 #ifdef HAVE_LOGIN_TTY
-				login_tty(x);
+					login_tty(x);
 
 #else
 				/* This tells the fd that it's a tty (I think) */
 #ifdef __svr4__
-				ioctl(x, I_PUSH, "ptem");
-				ioctl(x, I_PUSH, "ldterm");
+					joe_ioctl(x, I_PUSH, "ptem");
+					joe_ioctl(x, I_PUSH, "ldterm");
 #endif
 
 				/* Open stdout, stderr */
-				dup(x);
-				dup(x);	/* Standard output, standard error */
-				/* (yes, stdin, stdout, and stderr must all be open for reading and
-				 * writing.  On some systems the shell assumes this */
+					dup(x);
+					dup(x);	/* Standard output, standard error */
+					/* (yes, stdin, stdout, and stderr must all be open for reading and
+					 * writing.  On some systems the shell assumes this */
 #endif
 
-				/* We could probably have a special TTY set-up for JOE, but for now
-				 * we'll just use the TTY setup for the TTY was was run on */
 #ifdef HAVE_POSIX_TERMIOS
-				tcsetattr(0, TCSADRAIN, &oldterm);
+					tcsetattr(0, TCSADRAIN, &oldterm);
 #else
 #ifdef HAVE_SYSV_TERMIO
-				ioctl(0, TCSETAW, &oldterm);
+					joe_ioctl(0, TCSETAW, &oldterm);
 #else
-				ioctl(0, TIOCSETN, &oarg);
-				ioctl(0, TIOCSETC, &otarg);
-				ioctl(0, TIOCSLTC, &oltarg);
+					joe_ioctl(0, TIOCSETN, &oarg);
+					joe_ioctl(0, TIOCSETC, &otarg);
+					joe_ioctl(0, TIOCSLTC, &oltarg);
 #endif
 #endif
+					/* We could probably have a special TTY set-up for JOE, but for now
+					 * we'll just use the TTY setup for the TTY was was run on */
+
+					/* Execute the shell */
+					execve((char *)cmd, (char **)args, (char **)env);
+
+					/* If shell didn't execute */
+					joe_snprintf_1(buf,sizeof(buf),joe_gettext(_("Couldn't execute shell '%s'\n")),cmd);
+					write(1,(char *)buf,zlen(buf));
+					sleep(1);
+
+				} else {
+					unsigned char buf[1024];
+					int len;
+					dup(x); /* Standard error */
+					
+					for (;;) {
+						len = read(0, buf, sizeof(buf));
+						if (len > 0)
+							write(1, buf, len);
+						else
+							break;
+					}
+				}
 
-				/* Execute the shell */
-				execve((char *)cmd, (char **)args, (char **)env);
 
-				/* If shell didn't execute */
-				joe_snprintf_1((char *)buf,sizeof(buf),"Couldn't execute shell '%s'\n",cmd);
-				write(0,(char *)buf,zlen(buf));
-				sleep(1);
 			}
 
 			_exit(0);
