@@ -8,15 +8,12 @@
 #ifndef _JOE_UFORMAT_H
 #define _JOE_UFORMAT_H 1
 
-#include "config.h"
-#include "types.h"
-
 int ucenter PARAMS((BW *bw));
-P *pbop PARAMS((P *p));
-P *peop PARAMS((P *p));
+P *pbop PARAMS((BW *bw, P *p));
+P *peop PARAMS((BW *bw, P *p));
 int ubop PARAMS((BW *bw));
 int ueop PARAMS((BW *bw));
-void wrapword PARAMS((P *p, long int indent, int french, unsigned char *indents));
+void wrapword PARAMS((BW *bw, P *p, long int indent, int french, int no_over, unsigned char *indents));
 int uformat PARAMS((BW *bw));
 int ufmtblk PARAMS((BW *bw));
 
