@@ -5,36 +5,20 @@
  *
  *	This file is part of JOE (Joe's Own Editor)
  */
-#include "config.h"
 #include "types.h"
 
-#include <stdio.h>
-#include <math.h>
-#include <stdlib.h>
-
-#include "b.h"
-#include "pw.h"
-#include "utils.h"
-#include "vs.h"
-#include "utf8.h"
-#include "charmap.h"
-#include "ublock.h"
-#include "macro.h"
-#include "w.h"
-
 unsigned char *merr;
 
-extern int current_arg;
-extern int current_arg_set;
-
 int mode_hex;
 int mode_eng;
 int mode_ins;
 
+double vzero = 0.0;
+
 static RETSIGTYPE fperr(int unused)
 {
 	if (!merr) {
-		merr = US "Float point exception";
+		merr = joe_gettext(_("Float point exception"));
 	}
 	REINSTALL_SIGHANDLER(SIGFPE, fperr);
 }
@@ -94,7 +78,7 @@ static double expr(int prec, int en,struct var **rtv)
 		}
 		c = *ptr;
 		*ptr = 0;
-		if (!zcmp(s,US "joe")) {
+		if (!zcmp(s,USTR "joe")) {
 			*ptr = c;
 			v = 0;
 			x = 0.0;
@@ -108,7 +92,7 @@ static double expr(int prec, int en,struct var **rtv)
 					++q;
 				if (*q!=')') {
 					if (!merr)
-						merr = US "Missing )";
+						merr = joe_gettext(_("Missing )"));
 				} else
 					*q++ = 0;
 				if (en) {
@@ -119,71 +103,71 @@ static double expr(int prec, int en,struct var **rtv)
 						rmmacro(m);
 					} else {
 						if (!merr)
-							merr = US "Syntax error in macro";
+							merr = joe_gettext(_("Syntax error in macro"));
 					}
 				} else {
 					ptr = q;
 				}
 			} else {
 				if (!merr)
-					merr = US "Missing (";
+					merr = joe_gettext(_("Missing ("));
 			}
 			c = *ptr;
 		} else if (!en) {
 			v = 0;
 			x = 0.0;
-		} else if (!zcmp(s,US "hex")) {
+		} else if (!zcmp(s,USTR "hex")) {
 			mode_hex = 1;
 			mode_eng = 0;
-			v = get(US "ans");
+			v = get(USTR "ans");
 			x = v->val;
-		} else if (!zcmp(s,US "dec")) {
+		} else if (!zcmp(s,USTR "dec")) {
 			mode_hex = 0;
 			mode_eng = 0;
-			v = get(US "ans");
+			v = get(USTR "ans");
 			x = v->val;
-		} else if (!zcmp(s,US "eng")) {
+		} else if (!zcmp(s,USTR "eng")) {
 			mode_hex = 0;
 			mode_eng = 1;
-			v = get(US "ans");
+			v = get(USTR "ans");
 			x = v->val;
-		} else if (!zcmp(s,US "ins")) {
+		} else if (!zcmp(s,USTR "ins")) {
 			mode_ins = 1;
-			v = get(US "ans");
+			v = get(USTR "ans");
 			x = v->val;
-		} else if (!zcmp(s,US "sum")) {
+		} else if (!zcmp(s,USTR "sum")) {
 			double xsq;
 			int cnt = blksum(&x, &xsq);
 			if (!merr && cnt<=0)
-				merr = US "No numbers in block";
+				merr = joe_gettext(_("No numbers in block"));
 			v = 0;
-		} else if (!zcmp(s,US "cnt")) {
+		} else if (!zcmp(s,USTR "cnt")) {
 			double xsq;
 			int cnt = blksum(&x, &xsq);
 			if (!merr && cnt<=0)
-				merr = US "No numbers in block";
+				merr = joe_gettext(_("No numbers in block"));
 			v = 0;
 			x = cnt;
-		} else if (!zcmp(s,US "avg")) {
+		} else if (!zcmp(s,USTR "avg")) {
 			double xsq;
 			int cnt = blksum(&x, &xsq);
 			if (!merr && cnt<=0)
-				merr = US "No numbers in block";
+				merr = joe_gettext(_("No numbers in block"));
 			v = 0;
 			if (cnt)
 				x /= (double)cnt;
-		} else if (!zcmp(s,US "dev")) {
+		} else if (!zcmp(s,USTR "dev")) {
 			double xsq;
 			double avg;
 			int cnt = blksum(&x, &xsq);
 			if (!merr && cnt<=0)
-				merr = US "No numbers in block";
+				merr = joe_gettext(_("No numbers in block"));
 			v = 0;
 			if (cnt) {
 				avg = x / (double)cnt;
 				x = sqrt(xsq + (double)cnt*avg*avg - 2.0*avg*x);
 			}
-		} else if (!zcmp(s,US "eval")) {
+		} else if (!zcmp(s,USTR "eval")) {
 			unsigned char *save = ptr;
 			unsigned char *e = blkget();
 			if (e) {
@@ -192,7 +176,7 @@ static double expr(int prec, int en,struct var **rtv)
 				joe_free(e);
 				ptr = save;
 			} else if (!merr) {
-				merr = US "No block";
+				merr = joe_gettext(_("No block"));
 			}
 		} else {
 			v = get(s);
@@ -200,7 +184,9 @@ static double expr(int prec, int en,struct var **rtv)
 		}
 		*ptr = c;
 	} else if ((*ptr >= '0' && *ptr <= '9') || *ptr == '.') {
-		x = strtod((char *)ptr,(char **)&ptr);
+		char *eptr;
+		x = strtod((char *)ptr,&eptr);
+		ptr = (unsigned char *)eptr;
 	} else if (*ptr == '(') {
 		++ptr;
 		x = expr(0, en, &v);
@@ -208,7 +194,7 @@ static double expr(int prec, int en,struct var **rtv)
 			++ptr;
 		else {
 			if (!merr)
-				merr = US "Missing )";
+				merr = joe_gettext(_("Missing )"));
 		}
 	} else if (*ptr == '-') {
 		++ptr;
@@ -227,13 +213,13 @@ static double expr(int prec, int en,struct var **rtv)
 			++ptr;
 		else {
 			if (!merr)
-				merr = US "Missing )";
+				merr = joe_gettext(_("Missing )"));
 		}
 		if (v && v->func)
 			x = v->func(y);
 		else {
 			if (!merr)
-				merr = US "Called object is not a function";
+				merr = joe_gettext(_("Called object is not a function"));
 		}
 		goto loop;
 	} else if (*ptr == '!' && ptr[1]!='=' && 10 >= prec) {
@@ -247,7 +233,7 @@ static double expr(int prec, int en,struct var **rtv)
 			x = y;
 		} else {
 			if (!merr)
-				merr = US "Factorial can only take positive integers";
+				merr = joe_gettext(_("Factorial can only take positive integers"));
 		}
 		v = 0;
 		goto loop;
@@ -274,7 +260,7 @@ static double expr(int prec, int en,struct var **rtv)
 	} else if(*ptr=='%' && 7>prec) {
 		++ptr;
 		y = expr(7, en, &dumb);
-		if ((int)y == 0) x = 1.0/0.0;
+		if ((int)y == 0) x = 1.0/vzero;
 		else x = ((int) x) % (int)y;
 		v = 0;
 		goto loop;
@@ -334,7 +320,7 @@ static double expr(int prec, int en,struct var **rtv)
 				x = z;
 			v = 0;  
 		} else if (!merr) {
-			merr = US ": missing after ?";
+			merr = USTR ": missing after ?";
 		}
 		goto loop;
 	} else if (*ptr == '=' && 1 >= prec) {
@@ -345,7 +331,7 @@ static double expr(int prec, int en,struct var **rtv)
 			v->set = 1;
 		} else {
 			if (!merr)
-				merr = US "Left side of = is not an l-value";
+				merr = joe_gettext(_("Left side of = is not an l-value"));
 		}
 		v = 0;
 		goto loop;
@@ -356,17 +342,17 @@ static double expr(int prec, int en,struct var **rtv)
 
 static double eval(unsigned char *s)
 {
-	double result;
+	double result = 0.0;
 	struct var *v;
 	if(++recur==1000) {
-		merr = US "Recursion depth exceeded";
+		merr = joe_gettext(_("Recursion depth exceeded"));
 		--recur;
 		return 0.0;
 	}
 	ptr = s;
 	while (!merr && *ptr) {
 		result = expr(0, 1, &dumb);
-		v = get(US "ans");
+		v = get(USTR "ans");
 		v->val = result;
 		v->set = 1;
 		if (!merr) {
@@ -379,7 +365,7 @@ static double eval(unsigned char *s)
 					++ptr;
 				}
 			} else if (*ptr) {
-				merr = US "Extra junk after end of expr";
+				merr = joe_gettext(_("Extra junk after end of expr"));
 			}
 		}
 	}
@@ -423,77 +409,77 @@ double calc(BW *bw, unsigned char *s)
 	int c = brch(bw->cursor);
 
 	if (!vars) {
-		v = get(US "sin"); v->func = m_sin;
-		v = get(US "cos"); v->func = m_cos;
-		v = get(US "tan"); v->func = m_tan;
-		v = get(US "exp"); v->func = m_exp;
-		v = get(US "sqrt"); v->func = m_sqrt;
-		v = get(US "cbrt"); v->func = m_cbrt;
-		v = get(US "ln"); v->func = m_log;
-		v = get(US "log"); v->func = m_log10;
-		v = get(US "asin"); v->func = m_asin;
-		v = get(US "acos"); v->func = m_acos;
-		v = get(US "atan"); v->func = m_atan;
-		v = get(US "pi"); v->val = M_PI; v->set = 1;
-		v = get(US "e"); v->val = M_E; v->set = 1;
-		v = get(US "sinh"); v->func = m_sinh;
-		v = get(US "cosh"); v->func = m_cosh;
-		v = get(US "tanh"); v->func = m_tanh;
-		v = get(US "asinh"); v->func = m_asinh;
-		v = get(US "acosh"); v->func = m_acosh;
-		v = get(US "atanh"); v->func = m_atanh;
-		v = get(US "int"); v->func = m_int;
-		v = get(US "floor"); v->func = m_floor;
-		v = get(US "ceil"); v->func = m_ceil;
-		v = get(US "abs"); v->func = m_fabs;
-		v = get(US "erf"); v->func = m_erf;
-		v = get(US "erfc"); v->func = m_erfc;
-		v = get(US "j0"); v->func = m_j0;
-		v = get(US "j1"); v->func = m_j1;
-		v = get(US "y0"); v->func = m_y0;
-		v = get(US "y1"); v->func = m_y1;
+		v = get(USTR "sin"); v->func = m_sin;
+		v = get(USTR "cos"); v->func = m_cos;
+		v = get(USTR "tan"); v->func = m_tan;
+		v = get(USTR "exp"); v->func = m_exp;
+		v = get(USTR "sqrt"); v->func = m_sqrt;
+		v = get(USTR "cbrt"); v->func = m_cbrt;
+		v = get(USTR "ln"); v->func = m_log;
+		v = get(USTR "log"); v->func = m_log10;
+		v = get(USTR "asin"); v->func = m_asin;
+		v = get(USTR "acos"); v->func = m_acos;
+		v = get(USTR "atan"); v->func = m_atan;
+		v = get(USTR "pi"); v->val = M_PI; v->set = 1;
+		v = get(USTR "e"); v->val = M_E; v->set = 1;
+		v = get(USTR "sinh"); v->func = m_sinh;
+		v = get(USTR "cosh"); v->func = m_cosh;
+		v = get(USTR "tanh"); v->func = m_tanh;
+		v = get(USTR "asinh"); v->func = m_asinh;
+		v = get(USTR "acosh"); v->func = m_acosh;
+		v = get(USTR "atanh"); v->func = m_atanh;
+		v = get(USTR "int"); v->func = m_int;
+		v = get(USTR "floor"); v->func = m_floor;
+		v = get(USTR "ceil"); v->func = m_ceil;
+		v = get(USTR "abs"); v->func = m_fabs;
+		v = get(USTR "erf"); v->func = m_erf;
+		v = get(USTR "erfc"); v->func = m_erfc;
+		v = get(USTR "j0"); v->func = m_j0;
+		v = get(USTR "j1"); v->func = m_j1;
+		v = get(USTR "y0"); v->func = m_y0;
+		v = get(USTR "y1"); v->func = m_y1;
 	}
 
-	v = get(US "top");
+	v = get(USTR "top");
 	v->val = tbw->top->line + 1;
 	v->set = 1;
-	v = get(US "lines");
+	v = get(USTR "lines");
 	v->val = tbw->b->eof->line + 1;
 	v->set = 1;
-	v = get(US "line");
+	v = get(USTR "line");
 	v->val = tbw->cursor->line + 1;
 	v->set = 1;
-	v = get(US "col");
+	v = get(USTR "col");
 	v->val = tbw->cursor->col + 1;
 	v->set = 1;
-	v = get(US "byte");
+	v = get(USTR "byte");
 	v->val = tbw->cursor->byte + 1;
 	v->set = 1;
-	v = get(US "size");
+	v = get(USTR "size");
 	v->val = tbw->b->eof->byte;
 	v->set = 1;
-	v = get(US "height");
+	v = get(USTR "height");
 	v->val = tbw->h;
 	v->set = 1;
-	v = get(US "width");
+	v = get(USTR "width");
 	v->val = tbw->w;
 	v->set = 1;
-	v = get(US "char");
+	v = get(USTR "char");
 	v->val = (c == NO_MORE_DATA ? -1.0 : c);
 	v->set = 1;
-	v = get(US "markv");
+	v = get(USTR "markv");
 	v->val = markv(1) ? 1.0 : 0.0;
 	v->set = 1;
-	v = get(US "rdonly");
+	v = get(USTR "rdonly");
 	v->val = tbw->b->rdonly;
 	v->set = 1;
-	v = get(US "arg");
+	v = get(USTR "arg");
 	v->val = current_arg;
 	v->set = 1;
-	v = get(US "argset");
+	v = get(USTR "argset");
 	v->val = current_arg_set;
 	v->set = 1;
-	v = get(US "no_windows");
+	v = get(USTR "no_windows");
 	v->val = countmain(bw->parent->t);
 	v->set = 1;
 	merr = 0;
@@ -514,11 +500,11 @@ static int domath(BW *bw, unsigned char *s, void *object, int *notify)
 	}
 	vsrm(s);
 	if (mode_hex)
-		joe_snprintf_1((char *)msgbuf, JOE_MSGBUFSIZE, "0x%lX", (long)result);
+		joe_snprintf_1(msgbuf, JOE_MSGBUFSIZE, "0x%lX", (long)result);
 	else if (mode_eng)
-		joe_snprintf_1((char *)msgbuf, JOE_MSGBUFSIZE, "%.16G", result);
+		joe_snprintf_1(msgbuf, JOE_MSGBUFSIZE, "%.16G", result);
 	else
-		joe_snprintf_1((char *)msgbuf, JOE_MSGBUFSIZE, "%.16G", result);
+		joe_snprintf_1(msgbuf, JOE_MSGBUFSIZE, "%.16G", result);
 	if (bw->parent->watom->what != TYPETW || mode_ins) {
 		binsm(bw->cursor, sz(msgbuf));
 		pfwrd(bw->cursor, zlen(msgbuf));
@@ -535,7 +521,7 @@ B *mathhist = NULL;
 int umath(BW *bw)
 {
 	joe_set_signal(SIGFPE, fperr);
-	if (wmkpw(bw->parent, US "=", &mathhist, domath, US "Math", NULL, NULL, NULL, NULL, locale_map, 0)) {
+	if (wmkpw(bw->parent, USTR "=", &mathhist, domath, USTR "Math", NULL, NULL, NULL, NULL, locale_map, 0)) {
 		return 0;
 	} else {
 		return -1;
