@@ -8,6 +8,9 @@
  */
 #include "types.h"
 
+#ifdef HAVE_SYS_IOCTL_H
+#include <sys/ioctl.h>
+#endif
 
 #if 0
 int joe_ispunct(int wide,struct charmap *map,int c)
@@ -111,6 +114,15 @@ ssize_t joe_write(int fd, void *buf, size_t size)
 	return rt;
 }
 
+int joe_ioctl(int fd, int req, void *ptr)
+{
+	int rt;
+	do {
+		rt = ioctl(fd, req, ptr);
+	} while (rt == -1 && errno == EINTR);
+	return rt;
+}
+
 /* Heap checking versions of malloc() */
 
 /* #define HEAP_CHECK 1 */
@@ -241,17 +253,26 @@ void joe_free(void *ptr)
 
 void *joe_malloc(size_t size)
 {
-	return malloc(size);
+	void *p = malloc(size);
+	if (!p)
+		ttsig(-1);
+	return p;
 }
 
 void *joe_calloc(size_t nmemb,size_t size)
 {
-	return calloc(nmemb,size);
+	void *p = calloc(nmemb, size);
+	if (!p)
+		ttsig(-1);
+	return p;
 }
 
 void *joe_realloc(void *ptr,size_t size)
 {
-	return realloc(ptr,size);
+	void *p = realloc(ptr, size);
+	if (!p)
+		ttsig(-1);
+	return p;
 }
 
 void joe_free(void *ptr)
